"""\
socranop/contributors.py - autogenerated from CONTRIBUTORS.md

DO NOT EDIT THIS FILE. EDIT `CONTRIBUTORS.md` INSTEAD
and update this file by re-running contrib_to_about.
"""

authors = [
    "Jim Ramsay <i.am@jimramsay.com> - Author",
    "Christoph <soffioalcuore@posteo.net> - Testing and suggestions",
    "Pete Merges <pdmerges@gmail.com> - Notepad-8FX support and testing",
    "Viktor Mastoridis <viktor.mastoridis@gmail.com> - Notepad-5 support and testing",
    "Hans Ulrich Niedermann <hun@n-dimensional.de> - System integration",
]
artists = ["Flat Icons https://www.flaticon.com/authors/flat-icons"]
